import os
from urllib.parse import urlparse

from OpenSSL import crypto

from magicproxy.config import (
    PRIVATE_KEY_LOCATION,
    PUBLIC_KEY_LOCATION,
    PUBLIC_CERTIFICATE_LOCATION,
    PUBLIC_ACCESS,
)


def generate_keys(url=PUBLIC_ACCESS):
    os.makedirs(os.path.dirname(PRIVATE_KEY_LOCATION), exist_ok=True)
    os.makedirs(os.path.dirname(PUBLIC_KEY_LOCATION), exist_ok=True)
    os.makedirs(os.path.dirname(PUBLIC_CERTIFICATE_LOCATION), exist_ok=True)
    parsed = urlparse(url)
    if not parsed.hostname:
        raise ValueError(f"url {url} does not seem to have a hostname")

    hostname = str(parsed.hostname)
    pkey = crypto.PKey()
    pkey.generate_key(crypto.TYPE_RSA, 2048)

    with open(PRIVATE_KEY_LOCATION, "wb") as private_key_file:
        private_key_file.write(crypto.dump_privatekey(crypto.FILETYPE_PEM, pkey))

    with open(PUBLIC_KEY_LOCATION, "wb") as public_key_file:
        public_key_file.write(crypto.dump_publickey(crypto.FILETYPE_PEM, pkey))

    req = crypto.X509Req()
    subject = req.get_subject()
    setattr(subject, "CN", hostname)
    req.set_pubkey(pkey)
    req.sign(pkey, "sha256")

    certificate = crypto.X509()
    certificate.gmtime_adj_notBefore(0)
    certificate.gmtime_adj_notAfter(1825 * 24 * 60 * 60)
    certificate.set_subject(req.get_subject())
    certificate.set_issuer(req.get_subject())
    certificate.set_pubkey(pkey)
    certificate.sign(pkey, "sha256")

    with open(PUBLIC_CERTIFICATE_LOCATION, "wb") as certificate_file:
        certificate_file.write(
            crypto.dump_certificate(crypto.FILETYPE_PEM, certificate)
        )



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from urllib.parse import parse_qsl, urlencode, urlparse, urlunparse


def clean_path_queries(query_params_to_clean, path) -> str:
    parts = urlparse(path)
    if parts.query != "":
        queries = parse_qsl(parts.query, keep_blank_values=True, strict_parsing=True)
        cln = [q for q in queries if q[0] not in query_params_to_clean]
        return urlunparse(
            (
                str(parts.scheme),
                parts.netloc,
                parts.path,
                parts.params,
                urlencode(cln),
                parts.fragment,
            )
        )
    return path



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import re
import types
from typing import List, Union

from magicproxy.config import SCOPES
from magicproxy.types import Scope

logger = logging.getLogger(__name__)


def is_request_allowed(scope: Scope, method, path):
    logger.debug(f"validating request {method} {path} on scope {scope}")

    if method != scope.method and scope.method != "*":
        return False

    if not path.startswith("/"):
        path = f"/{path}"

    if re.match(scope.path, path, re.I):
        return True


def validate_request(
    method: str,
    path: str,
    scopes: List[str] = None,
    allowed: List[str] = None,
) -> bool:
    """Basic scope validation routine.

    Args:
        method: The request HTTP method.
        path: The request path.
        scopes: The list of allowed named scopes.
        allowed: The list of allowed requests.

    The named scope must be configured in the proxy
    Or allowed passed through a METHOD path string like this:

        METHOD path

    For example:

        GET /user
        POST /repos/+?/+?/issues/+?/labels

    Would allow getting the user info and updating labels on issues on a GitHub repo
    """
    if allowed is None:
        allowed = []

    if scopes is None:
        scopes = []

    for scope_key in scopes:
        scope_element: Union[List[Scope], types.ModuleType] = SCOPES[scope_key]
        if isinstance(scope_element, list):
            for scope in scope_element:
                if is_request_allowed(scope, method, path):
                    return True
        elif isinstance(scope_element, types.ModuleType):
            if hasattr(scope_element, "is_request_allowed"):
                return scope_element.is_request_allowed(method=method, path=path)

    for allowed_item in allowed:
        allowed_method, allowed_path = allowed_item.split(" ", 1)
        if is_request_allowed(
            Scope(method=allowed_method, path=allowed_path), method, path
        ):
            return True

    return False


def response_callback(content, code, headers, scopes=None):
    """Response callback, for dynamic proxies

    Args:
        scopes: The list of allowed named scopes.

    The named scope must be configured in the proxy

    Would allow the proxy to process the response of a request
    e.g. can allow a DELETE on a resource once its :id: is known in the JSON of the reponse
    """
    if scopes is None:
        scopes = []

    for scope_key in scopes:
        scope_element: Union[List[Scope], types.ModuleType] = SCOPES[scope_key]
        if isinstance(scope_element, types.ModuleType):
            if hasattr(scope_element, "response_callback"):
                return scope_element.response_callback(
                    content=content, code=code, headers=headers
                )



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


DEFAULT_REMOVED_REQUEST_HEADERS = {"Host", "Connection", "Authorization"}

DEFAULT_REMOVED_RESPONSE_HEADERS = {
    "Content-Length",
    "Content-Encoding",
    "Transfer-Encoding",
}


def clean_request_headers(headers, custom_clean_headers):
    """Removes HTTP Headers for a Request

    Args:
      headers: the HTTP headers of the request
      custom_clean_headers: a list of additional headers to remove

    Returns:
      HTTP headers that have been cleaned of unwanted values
    """
    headers = dict(headers)
    for rmv in DEFAULT_REMOVED_REQUEST_HEADERS.union(custom_clean_headers):
        headers.pop(rmv, None)
    return headers


def clean_response_headers(headers):
    """Removes HTTP Headers for a Response

    Args:
      headers: the HTTP headers of the response

    Returns:
      HTTP headers that have been cleaned of unwanted values
    """
    headers = dict(headers)
    for rmv in DEFAULT_REMOVED_RESPONSE_HEADERS:
        headers.pop(rmv, None)
    headers["X-Thea-Codes-GitHub-Proxy"] = "1.1"
    return headers



import glob
import logging
import os
import traceback
from importlib import util
import inspect

logger = logging.getLogger()


class InvalidPluginError(Exception):
    pass


class PluginNotFoundError(Exception):
    pass


def load_module(path):
    name = os.path.split(path)[-1]
    spec = util.spec_from_file_location(name, path)
    module = util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module


def load_plugins(plugins_folder):
    plugins = {}
    for python_file in glob.glob(f"{plugins_folder}/*.py"):
        logging.debug("load_plugin %s", python_file)
        scope_key, module = load_plugin(python_file)
        plugins[scope_key] = module

    return plugins


def load_plugin(python_file):
    scope_key = os.path.splitext(os.path.basename(python_file))[0]
    plugin_str = f"{scope_key} ({python_file})"
    if not os.path.exists(python_file):
        raise PluginNotFoundError("this plugin file does not exist")
    try:
        module = load_module(python_file)
    except Exception as e:
        logger.error("%s not importable", plugin_str)
        logger.error(traceback.format_exc())
        raise InvalidPluginError() from e

    has_is_requests_allowed = hasattr(module, "is_request_allowed")
    has_response_callback = hasattr(module, "response_callback")

    if has_is_requests_allowed or has_response_callback:
        if has_is_requests_allowed:
            signature = inspect.signature(module.is_request_allowed)
            if (
                "method" not in signature.parameters
                and "path" not in signature.parameters
            ):
                raise InvalidPluginError(
                    "%s is_request_allowed member needs 'method', 'path' parameters",
                    plugin_str,
                )

        if has_response_callback:
            signature = inspect.signature(module.response_callback)
            if (
                "content" not in signature.parameters
                and "code" not in signature.parameters
                and "headers" not in signature.parameters
            ):
                raise InvalidPluginError(
                    "%s response_callback member needs 'content', 'code', 'headers' parameters",
                    plugin_str,
                )
    else:
        raise InvalidPluginError(
            "%s no member is_request_allowed or request_callback", plugin_str
        )

    return scope_key, module



from dataclasses import dataclass
from typing import Optional, List, Union

import google.auth
import google.auth.crypt
import google.auth.jwt
from cryptography import x509
from cryptography.hazmat.primitives.asymmetric import rsa


@dataclass
class Scope:
    method: str
    path: str


@dataclass
class DecodeResult:
    token: str
    scopes: Optional[List[Union[str, Scope]]]
    allowed: Optional[List[Union[str, Scope]]]


@dataclass
class _Keys:
    private_key: rsa.RSAPrivateKey = None
    private_key_signer: google.auth.crypt.RSASigner = None
    public_key: rsa.RSAPublicKey = None
    certificate: x509.Certificate = None
    certificate_pem: bytes = None



import json
import logging
import os
from collections.abc import Mapping
from typing import Union

from magicproxy.plugins import load_plugins
from magicproxy.types import Scope

logger = logging.getLogger(__name__)

DEFAULT_API_ROOT = "https://api.github.com"
DEFAULT_PRIVATE_KEY_LOCATION = "keys/private.pem"
DEFAULT_PUBLIC_KEY_LOCATION = "keys/public.pem"
DEFAULT_PUBLIC_CERTIFICATE_LOCATION = "keys/public.x509.cer"
DEFAULT_PUBLIC_ACCESS = "http://localhost"

CONFIG_FILE = os.environ.get("CONFIG_FILE")

config = {}

if CONFIG_FILE is not None:
    try:
        config_string = open(CONFIG_FILE, "r", encoding="utf-8").read()
    except IOError:
        raise RuntimeError("I/O error, config file should be readable")
    try:
        config = json.loads(config_string)
    except ValueError:
        raise RuntimeError("config file should be a valid JSON file")

logger.debug("config %s", config)
PLUGINS_LOCATION = config.get("plugins_location")

API_ROOT = os.environ.get("API_ROOT", config.get("api_root", DEFAULT_API_ROOT))
PRIVATE_KEY_LOCATION = os.environ.get(
    "PRIVATE_KEY_LOCATION",
    config.get("private_key_location", DEFAULT_PRIVATE_KEY_LOCATION),
)
PUBLIC_KEY_LOCATION = os.environ.get(
    "PUBLIC_KEY_LOCATION",
    config.get("public_key_location", DEFAULT_PUBLIC_KEY_LOCATION),
)
PUBLIC_CERTIFICATE_LOCATION = os.environ.get(
    "PUBLIC_CERTIFICATE_LOCATION",
    config.get("public_certificate_location", DEFAULT_PUBLIC_CERTIFICATE_LOCATION),
)
PUBLIC_ACCESS = os.environ.get(
    "PUBLIC_ACCESS", config.get("public_access", DEFAULT_PUBLIC_ACCESS)
)

SCOPES = config.get("scopes", {})


def parse_scope(element: Union[str, Mapping]) -> Scope:
    logging.debug("parsing scope from %s", element)
    if isinstance(element, str):
        try:
            method, path = element.split(" ", 1)
            return Scope(method=method, path=path)
        except ValueError as e:
            raise ValueError('a scope string should be a "METHOD path_regex"') from e
    elif isinstance(element, Mapping):
        if "method" in element and "path" in element:
            return Scope(method=element["method"], path=element["path"])
        else:
            raise ValueError(
                "a scope mapping should be a mapping with method, path keys"
            )


for scope_key in SCOPES:
    scope_elements = []
    for scope_element in SCOPES[scope_key]:
        scope_elements.append(parse_scope(scope_element))
    SCOPES[scope_key] = scope_elements

logger.debug("PLUGINS_LOCATION %s", PLUGINS_LOCATION)
if PLUGINS_LOCATION:
    SCOPES.update(**load_plugins(PLUGINS_LOCATION))

logger.debug("SCOPES %s", SCOPES)



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import traceback
from typing import Tuple, Set

import flask
import requests

import magicproxy
import magicproxy.types
from .config import API_ROOT
from .headers import clean_request_headers, clean_response_headers
from . import magictoken
from . import scopes
from . import queries
from .magictoken import magictoken_params_validate

logger = logging.getLogger(__name__)

app = flask.Flask(__name__)

query_params_to_clean: Set[str] = set()

custom_request_headers_to_clean: Set[str] = set()


@app.route("/__magictoken", methods=["POST", "GET"])
def create_magic_token():
    if flask.request.method == "GET":
        return "magic API proxy for " + API_ROOT + " version " + magicproxy.__version__
    params = flask.request.json
    try:
        magictoken_params_validate(params)
    except ValueError as e:
        return str(e), 400

    token = magictoken.create(
        keys, params["token"], params.get("scopes"), params.get("allowed")
    )

    return token, 200, {"Content-Type": "application/jwt"}


def _proxy_request(
    request: flask.Request, url: str, headers=None, **kwargs
) -> Tuple[bytes, int, dict]:
    clean_headers = clean_request_headers(
        request.headers, custom_request_headers_to_clean
    )

    if headers:
        clean_headers.update(headers)

    print(
        f"Proxying to {request.method} {url}\nHeaders: {clean_headers}\nQuery: {request.args}\nContent: {request.data!r}"
    )

    # Make the API request
    resp = requests.request(
        url=url,
        method=request.method,
        headers=clean_headers,
        params=dict(request.args),
        data=request.data,
        **kwargs,
    )

    response_headers = clean_response_headers(resp.headers)

    print(resp, resp.headers, resp.content)

    return resp.content, resp.status_code, response_headers


@app.route("/", defaults={"path": ""})
@app.route("/<path:path>", methods=["POST", "GET", "PATCH", "PUT", "DELETE"])
def proxy_api(path):
    auth_token = flask.request.headers.get("Authorization")
    if auth_token is None:
        return "No authorization token presented", 401
    # strip out "Bearer " if needed
    if auth_token.startswith("Bearer "):
        auth_token = auth_token[len("Bearer ") :]

    try:
        # Validate the magic token
        token_info = magictoken.decode(keys, auth_token)
    except ValueError:
        return "Not a valid magic token", 400

    # Validate scopes against URL and method.
    if not scopes.validate_request(
        flask.request.method, path, token_info.scopes, token_info.allowed
    ):
        return (
            f"Disallowed by API proxy. Allowed scopes: {', '.join(token_info.scopes)}",
            401,
        )

    path = queries.clean_path_queries(query_params_to_clean, path)

    response = _proxy_request(
        request=flask.request,
        url=f"{API_ROOT}/{path}",
        headers={"Authorization": f"Bearer {token_info.token}"},
    )

    try:
        scopes.response_callback(*response, token_info.scopes)
    except Exception as e:
        logger.error("exception in response_callback")
        logger.error(e)
        logger.error(traceback.format_exc())

    return response


def run_app(host, port):
    global keys
    keys = magictoken.Keys.from_env()
    app.run(host=host, port=port)



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import calendar
import datetime

import google.auth.crypt
import google.auth.jwt
from cryptography import x509
from cryptography.hazmat import backends
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import padding

from magicproxy.config import PRIVATE_KEY_LOCATION, PUBLIC_CERTIFICATE_LOCATION, SCOPES
from magicproxy.crypto import generate_keys
from magicproxy.types import DecodeResult, _Keys

VALIDITY_PERIOD = 365 * 5  # 5 years.

_BACKEND = backends.default_backend()
_PADDING = padding.OAEP(
    mgf=padding.MGF1(algorithm=hashes.SHA256()), algorithm=hashes.SHA256(), label=None
)


def _datetime_to_secs(value: datetime.datetime) -> int:
    return calendar.timegm(value.utctimetuple())


def _encrypt(key, plain_text: bytes) -> bytes:
    return key.encrypt(plain_text, _PADDING)


def _decrypt(key, cipher_text: bytes) -> bytes:
    return key.decrypt(
        cipher_text,
        padding.OAEP(
            mgf=padding.MGF1(algorithm=hashes.SHA256()),
            algorithm=hashes.SHA256(),
            label=None,
        ),
    )


class Keys(_Keys):
    @classmethod
    def from_files(cls, private_key_file, certificate_file):
        with open(private_key_file, "rb") as fh:
            private_key_bytes = fh.read()
            private_key = serialization.load_pem_private_key(
                private_key_bytes, password=None, backend=_BACKEND
            )
            private_key_signer = google.auth.crypt.RSASigner.from_string(
                private_key_bytes
            )

        with open(certificate_file, "rb") as fh:
            certificate_pem = fh.read()
            certificate = x509.load_pem_x509_certificate(certificate_pem, _BACKEND)
            public_key = certificate.public_key()

        return cls(
            private_key=private_key,
            private_key_signer=private_key_signer,
            public_key=public_key,
            certificate=certificate,
            certificate_pem=certificate_pem,
        )

    @classmethod
    def from_env(cls):
        try:
            return Keys.from_files(PRIVATE_KEY_LOCATION, PUBLIC_CERTIFICATE_LOCATION)
        except FileNotFoundError:
            generate_keys()
            return Keys.from_files(PRIVATE_KEY_LOCATION, PUBLIC_CERTIFICATE_LOCATION)


def create(keys: _Keys, token, scopes=None, allowed=None) -> str:
    # NOTE: This is the *public key* that we use to encrypt this token. It's
    # *extremely* important that the public key is used here, as we want only
    # our *private key* to be able to decrypt this value.
    encrypted_api_token = _encrypt(keys.public_key, token.encode("utf-8"))
    encoded_api_token = base64.b64encode(encrypted_api_token).decode("utf-8")

    issued_at = datetime.datetime.utcnow()
    expires_at = datetime.datetime.utcnow() + datetime.timedelta(days=VALIDITY_PERIOD)

    claims = {
        "iat": _datetime_to_secs(issued_at),
        "exp": _datetime_to_secs(expires_at),
        "token": encoded_api_token,
    }

    if allowed:
        claims["allowed"] = allowed

    if scopes:
        claims["scopes"] = scopes

    jwt = google.auth.jwt.encode(keys.private_key_signer, claims)

    return jwt.decode("utf-8")


def decode(keys, token) -> DecodeResult:
    claims = dict(
        google.auth.jwt.decode(token, verify=True, certs=[keys.certificate_pem])
    )

    decoded_token = base64.b64decode(claims["token"])
    decrypted_token = _decrypt(keys.private_key, decoded_token).decode("utf-8")
    claims["token"] = decrypted_token

    return DecodeResult(claims["token"], claims.get("scopes"), claims.get("allowed"))


def magictoken_params_validate(params: dict):
    if not params:
        raise ValueError("Request must be json")

    if "token" not in params:
        raise ValueError("We need a token for the API behind, in the 'token' field")

    if "scopes" in params and "allowed" in params:
        raise ValueError(
            "allowed (spelling out the allowed requests) "
            "OR scopes (naming a scope configured on the proxy, not both"
        )

    if "scopes" in params:
        if not isinstance(params.get("scopes"), list):
            raise ValueError("scopes must be a list")
        params_scopes = params.get("scopes", [])
        if not all(isinstance(r, str) for r in params_scopes):
            raise ValueError("scopes must be a list of strings")
        if not all(r in SCOPES for r in params_scopes):
            raise ValueError(
                f"scopes must be configured on the proxy (valid: {' '.join(SCOPES)})"
            )

    elif "allowed" in params:
        if not isinstance(params.get("allowed"), list):
            raise ValueError("allowed must be a list of ")
        if not all(isinstance(r, str) for r in params.get("allowed", [])):
            raise ValueError("allowed must be a list of strings")
    else:
        raise ValueError(
            "need one of allowed (spelling out the allowed requests) "
            "OR scopes (naming a scope configured on the proxy)"
        )



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from setuptools import setup, find_packages
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="magic-api-proxy",
    description="A stateless API proxy",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/rienafairefr/magic-api-proxy",
    author="adapted from Alethea Katherine Flowers",
    author_email="matthieu@mmea.fr",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    keywords="api github proxy",
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    install_requires=["google-auth", "flask", "cryptography", "attrs", "requests", "aiohttp", "pyopenssl"],
    python_requires=">=3.6",
    project_urls={
        "Bug Reports": "https://github.com/rienafairefr/magic-api-proxy/issues",
        "Source": "https://github.com/rienafairefr/magic-api-proxy",
    },
)



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import argparse

import aiohttp.web

import logging

logger = logging.getLogger()
logging.basicConfig(level=logging.DEBUG)

parser = argparse.ArgumentParser(description="magicproxy server")
parser.add_argument(
    "--async",
    action="store_true",
    dest="run_async",
    help="run async using aiohttp (single process)",
)
parser.add_argument("--port", type=int, default=5000)
parser.add_argument("--host", type=str, default="127.0.0.1")


if __name__ == "__main__":
    from magicproxy import proxy, async_proxy

    args = parser.parse_args()
    if args.run_async:
        aiohttp.web.run_app(async_proxy.build_app([]), host=args.host, port=args.port)
    else:
        proxy.run_app(host=args.host, port=args.port)



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
from typing import Set

import aiohttp
import aiohttp.web

import magicproxy
from magicproxy.magictoken import Keys, magictoken_params_validate
from .config import API_ROOT
from .headers import clean_request_headers, clean_response_headers
from . import magictoken
from . import scopes
from . import queries

routes = aiohttp.web.RouteTableDef()
logger = logging.getLogger(__name__)

query_params_to_clean: Set[str] = set()
custom_request_headers_to_clean: Set[str] = set()


@routes.get("/__magictoken")
async def magic_token_version(request):
    return aiohttp.web.Response(
        body="magic API proxy for " + API_ROOT + " version " + magicproxy.__version__
    )


@routes.post("/__magictoken")
async def create_magic_token(request):
    params = await request.json()

    try:
        magictoken_params_validate(params)
    except ValueError as e:
        raise aiohttp.web.HTTPBadRequest(body=str(e))

    token = magictoken.create(
        keys, params["token"], params.get("scopes"), params.get("allowed")
    )

    return aiohttp.web.Response(body=token, headers={"Content-Type": "application/jwt"})


async def _proxy_request(request, url, headers=None, **kwargs):
    clean_headers = clean_request_headers(
        request.headers, custom_request_headers_to_clean
    )

    if headers:
        clean_headers.update(headers)

    print(f"Proxying to {request.method} {url}\n")

    async with aiohttp.ClientSession() as session:
        proxied_request = session.request(
            url=url,
            method=request.method,
            headers=clean_headers,
            params=request.query,
            data=request.content,
            **kwargs,
        )
        async with proxied_request as proxied_response:
            response_headers = clean_response_headers(proxied_response.headers)

            response = aiohttp.web.StreamResponse(
                status=proxied_response.status, headers=response_headers
            )

            await response.prepare(request)

            async for data, last in proxied_response.content.iter_chunks():
                await response.write(data)

            await response.write_eof()

            return data, proxied_response.status, proxied_response.headers


@routes.route("*", "/{path:.*}")
async def proxy_api(request):
    path = request.match_info["path"]

    auth_token = request.headers.get("Authorization")
    if auth_token is None:
        raise aiohttp.web.HTTPForbidden(body="No authorization token presented")

    # strip out "Bearer " if needed
    if auth_token.startswith("Bearer "):
        auth_token = auth_token[len("Bearer ") :]

    # Validate the magic token
    token_info = magictoken.decode(keys, auth_token)

    # Validate scopes againt URL and method.
    if not scopes.validate_request(
        request.method, request.path, token_info.scopes, token_info.allowed
    ):
        raise aiohttp.web.HTTPForbidden(
            body=f"Disallowed by API proxy. Allowed scopes: {', '.join(token_info.scopes)}"
        )

    path = queries.clean_path_queries(query_params_to_clean, path)

    response = await _proxy_request(
        request=request,
        url=f"{API_ROOT}/{path}",
        headers={"Authorization": f"Bearer {token_info.token}"},
    )

    try:
        scopes.response_callback(*response, token_info.scopes)
    except Exception as e:
        logger.error(e)
    return response


async def build_app(argv):
    global keys
    keys = Keys.from_env()

    app = aiohttp.web.Application()
    app.add_routes(routes)
    return app
